/* SAPER */
/* A game coded by 'koles' Adam Michalski */
/* (C) 1999 */

#include <bios.h>
#include <process.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <time.h>

extern unsigned _stklen = 16384;

char far *ScrBuffPtr = (char far *)0xB8000000L;

#define MINE -1
#define EMPTY 0

int width, height, mines, x_on_screen, y_on_screen, uncovered_fields,
    uncovered_mines;
int field[80][23]; //a po co sie bede klopotal z dynamiczna alokacja? :-)
int uncovered[80][23];
int marked[80][23];
int colors[9] = {0, 9, 2, 12, 1, 4, 3, 5, 15};

int getkey(void)
{
  int key, lo, hi;

  key = bioskey(0);
  lo = key & 0X00FF;
  hi = (key & 0XFF00) >> 8;
  return((lo == 0) ? hi + 256 : lo);
}

int MOUSE_Init(void)
{
  asm mov ax, 0
  asm int 0x33
  return _AX;
}

void MOUSE_Configure(void)
{
  asm mov ax, 7
  asm mov cx, 0
  asm mov dx, 79
  asm int 0x33

  asm mov ax, 8
  asm mov cx, 0
  asm mov dx, 24
  asm int 0x33
}

void MOUSE_ShowCursor(void)
{
  asm mov ax, 1
  asm int 0x33
}

void MOUSE_HideCursor(void)
{
  asm mov ax, 2
  asm int 0x33
}

void MOUSE_GetPos(int *x, int *y)
{
  asm mov ax, 3
  asm int 0x33
  *x = _CX >> 3;
  *y = _DX >> 3;
}

int MOUSE_GetButtonsState(void)
{
  asm mov ax, 5
  asm int 0x33
  return _AX;
}

void DispString(int Column, int Row, int Attr, const char *String, int Length)
{
  char far *Ptr = ScrBuffPtr + (Row * 80 + Column) * 2;
  while (Length--)
  {
    *Ptr++ = *String++;
    *Ptr++ = Attr;
  }
}

int DispWithFrmt(int Column, int Row, int Attr, const char *Format, ...)
{
  va_list ArgPtr;
  char Buffer[80];
  int Counter;

  va_start(ArgPtr, Format);
  Counter = vsprintf(Buffer, Format, ArgPtr);
  va_end(ArgPtr);

  DispString(Column, Row, Attr, Buffer, Counter);

  return Counter;
}

void InitPlayField(void)
{
  int xp, yp, xk, yk;
  textcolor(15);
  clrscr();
  do
  {
    printf("Podaj dlugosc: ");
    scanf("%d", &width);
  } while(width < 4 || width > 80);
  do
  {
    printf("Podaj szerokosc: ");
    scanf("%d", &height);
  } while(height < 4 || height > 23);
  do
  {
    printf("Podaj ilosc min: ");
    scanf("%d", &mines);
  } while (mines < 1 || mines > width*height);
  uncovered_mines = mines;
  clrscr();

  x_on_screen = (80 - width) / 2;
  y_on_screen = (23 - height) / 2;
  xk = x_on_screen + width - 1;
  yk = y_on_screen + height - 1;

  for (yp = y_on_screen ; yp <= yk; yp++)
    for (xp = x_on_screen; xp <= xk; xp++)
      DispWithFrmt(xp, yp, 7, "");

  MOUSE_Init();
  MOUSE_ShowCursor();
}

void FillField(void)
{
  int m = 0, x, y;
  randomize();
  while (m < mines)
  {
    do
    {
      x = random(width);
      y = random(height);
    } while (field[x][y] == MINE || x == 0 && y == 0);
    field[x][y] = MINE;
    m++;
  }

  //ok, teraz wypelniamy dane o sasiadach
  for (x = 0; x < width; x++)
  {
    for (y = 0; y < height; y++)
    {
      if (field[x][y] != MINE)
      {
        m = 0;
        if (x-1 >= 0 && y-1 >= 0 && field[x-1][y-1] == MINE) m++;
        if (y-1 >= 0 && field[x][y-1] == MINE) m++;
        if (x+1 < width && y-1 >= 0 && field[x+1][y-1] == MINE) m++;
        if (x-1 >= 0 && field[x-1][y] == MINE) m++;
        if (x+1 < width && field[x+1][y] == MINE) m++;
        if (x-1 >= 0 && y+1 < height && field[x-1][y+1] == MINE) m++;
        if (y+1 < height && field[x][y+1] == MINE) m++;
        if (x+1 < width && y+1 < height && field[x+1][y+1] == MINE) m++;
        if (m)
          field[x][y] = m;
      }
    }
  }
}

void UncvrField(int x, int y)
{
  if (x < 0 || x >= width || y < 0 || y >= height)
    return;
  if (uncovered[x][y])
    return;
  uncovered[x][y] = 1;
  uncovered_fields++;
  DispWithFrmt(x + x_on_screen, y + y_on_screen, colors[field[x][y]], "%c",
               field[x][y] == EMPTY ? ' ' : field[x][y] + '0');
  if (field[x][y] == EMPTY)
  {
    UncvrField(x-1, y-1);
    UncvrField(x, y-1);
    UncvrField(x+1, y-1);
    UncvrField(x-1, y);
    UncvrField(x+1, y);
    UncvrField(x-1, y+1);
    UncvrField(x, y+1);
    UncvrField(x+1, y+1);
  }
}

void UncoverField(int x, int y)
{
  if (field[x][y] == MINE)
  {
    DispWithFrmt(x + x_on_screen, y + y_on_screen, 16*7+4, "*");
    DispWithFrmt(0, 24, 4, "                                      ");
    DispWithFrmt(0, 24, 4, "Trafie na min kolego :-)");
    getkey();
    exit(0);
  }
  if (uncovered[x][y])
    return;
  UncvrField(x, y);
}

int main(void)
{
  int x, y, xn, yn;
  InitPlayField();
  FillField();
  DispWithFrmt(40, 23, 11, "Saper. Coded by Adam Michalski (aimsoft)");
  DispWithFrmt(0, 23, 14, "%5d", uncovered_mines);

  while(width*height - uncovered_fields != mines)
  {
    if (MOUSE_GetButtonsState() == 1)
    {
      MOUSE_GetPos(&x, &y);
      while (MOUSE_GetButtonsState() != 0)
        ;
      MOUSE_GetPos(&xn, &yn);
      if (x == xn && y == yn && x >= x_on_screen && x < x_on_screen + width
                             && y >= y_on_screen && y < y_on_screen + height
                             && !marked[x - x_on_screen][y - y_on_screen])
      {
        MOUSE_HideCursor();
        UncoverField(x - x_on_screen, y - y_on_screen);
        MOUSE_ShowCursor();
      }
    }
    else if (MOUSE_GetButtonsState() == 2)
    {
      MOUSE_GetPos(&x, &y);
      while (MOUSE_GetButtonsState() != 0)
        ;
      MOUSE_GetPos(&xn, &yn);
      if (x == xn && y == yn && x >= x_on_screen && x < x_on_screen + width
                             && y >= y_on_screen && y < y_on_screen + height
                             && !uncovered[x - x_on_screen][y - y_on_screen])
      {
        MOUSE_HideCursor();
        marked[x - x_on_screen][y - y_on_screen] ^= 1;
        if (marked[x - x_on_screen][y - y_on_screen])
          uncovered_mines--;
        else
          uncovered_mines++;
        DispWithFrmt(x, y, marked[x - x_on_screen][y - y_on_screen] ? 16*7 :
                     7, "%c", marked[x - x_on_screen][y - y_on_screen] ? '?'
                     : '');
        DispWithFrmt(0, 23, 14, "%5d", uncovered_mines);
        MOUSE_ShowCursor();
      }
    }
  }
  DispWithFrmt(0, 24, 4, "                                      ");
  DispWithFrmt(0, 24, 4, "Wygrae kolego - gratuluj :-)");
  getkey();

  return 0;
}
